unit GsScroller;

(*************************************************************)
(*                                                           *)
(*  TGsScroller (from G's Components Collection)             *)
(*                                                           *)
(*  Copyright  1999 - 2001 by Radu Buzea.                   *)
(*  Copyright  2001 by Rospotline Ltd.                      *)
(*  All rights reserved.                                     *)
(*                                                           *)
(*  Contacts at:                                             *)
(*  - radu@rospotline.com                                    *)
(*  - office@rospotline.com                                  *)
(*                                                           *)
(*************************************************************)
(*                                                           *)
(*  Permission to use, copy, modify, and distribute this     *)
(*  software without fee for any non commerical purpose      *)
(*  is hereby granted, provided that the above copyright     *)
(*  notice appears on all copies and that both that          *)
(*  copyright notice and this permission notice appear       *)
(*  in all supporting documentation.                         *)
(*                                                           *)
(*  NO REPRESENTATIONS ARE MADE ABOUT THE SUITABILITY OF     *)
(*  THIS SOFTWARE FOR ANY PURPOSE. THIS SOFTWARE IS          *)
(*  PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.    *)
(*  NEITHER Radu Buzea OR Rospotline Ltd. BE LIABLE FOR ANY  *)
(*  DAMAGES SUFFERED BY THE USE OF THIS SOFTWARE.            *)
(*                                                           *)
(*************************************************************)

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, ExtCtrls, ShellAPI;

type
   TGsScroller = class;

   TScrollDirection = (sdRightLeft, sdLeftRight);
   TScrollSpeed = (speed1_VerySlow, speed2_Slow, speed3_Normal, speed4_Fast, speed5_VeryFast);
   TScrollerOption = (opActiveMouse, opDirectDraw, opDrawFocus, opExecuteHyperlinks);
   TScrollerOptions = set of TScrollerOption;

   TScrollerItem = class(TCollectionItem)
   private
      FImg, FTextImg: TBitmap;
      FText1, FText2: TCaption;
      FHyperlink: string;
      FAlign: TAlignment;
      FUseText: Boolean;
      FMargin: Word;
      FBkColor: TColor;
      FFont: TFont;
      FOnChange: TNotifyEvent;
      procedure SetImg(Value: TBitmap);
      procedure SetText1(Value: TCaption);
      procedure SetText2(Value: TCaption);
      procedure SetAlign(Value: TAlignment);
      procedure SetUseText(Value: Boolean);
      procedure SetMargin(Value: Word);
      procedure SetBkColor(Value: TColor);
   protected
      procedure DrawTextBitmap;
      procedure FontChange(Sender: TObject);
   public
      constructor Create(Collection: TCollection); override;
      destructor Destroy; override;
      procedure Assign(Source: TPersistent); override;
      function GetFace: TBitmap;
      procedure Repaint;
      procedure Change; dynamic;
   published
      property Alignment: TAlignment read FAlign write SetAlign;
      property Background: TColor read FBkColor write SetBkColor;
      property Bitmap: TBitmap read FImg write SetImg;
      property Font: TFont read FFont write FFont;
      property Hyperlink: string read FHyperlink write FHyperlink;
      property Margin: Word read FMargin write SetMargin;
      property UseTextImage: Boolean read FUseText write SetUseText;
      property Text1: TCaption read FText1 write SetText1;
      property Text2: TCaption read FText2 write SetText2;
      property OnChange: TNotifyEvent read FOnChange write FOnChange;
   end;

   TScrollerCollection = class(TCollection)
   private
      FOwner: TGsScroller;
      function GetItem(Index: Integer): TScrollerItem;
      procedure SetItem(Index: Integer; Value: TScrollerItem);
   protected
      function GetOwner: TPersistent; override;
   public
      constructor Create(Owner: TGsScroller);
      function Add: TScrollerItem;
      property Items[Index: Integer]: TScrollerItem read GetItem write SetItem; default;
   end;

   THyperlinkClick = procedure(Sender: TObject; var Hyperlink: string;
     var Execute: Boolean) of Object;

   TGsScroller = class(TCustomControl)
   private
      FItems: TScrollerCollection;
      FTimer: TTimer;
      FBorderStyle: TBorderStyle;
      FDirection: TScrollDirection;
      IncLeft, IncRight, LastLeft, LastRight: Integer;
      FMouseX, FMouseItm, FMouseItmLeft, FMouseDifX: Integer;
      FSpeed: TScrollSpeed;
      FIsMouseDown: Boolean;
      FOptions: TScrollerOptions;
      FCursor, FCursorHold: TCursor;
      FOnHyperlinkClick: THyperlinkClick;
      procedure SetBorderStyle(Value: TBorderStyle);
      procedure SetDirection(Value: TScrollDirection);
      function GetActive: Boolean;
      procedure SetActive(Value: Boolean);
      procedure SetSpeed(Value: TScrollSpeed);
      procedure SetItems(Value: TScrollerCollection);
      procedure SetOptions(Value: TScrollerOptions);
   protected
      procedure Paint; override;
      procedure Resize; override;
      procedure ScrollTime(Sender: TObject);
      procedure CreateParams(var Params: TCreateParams); override;
      procedure DrawLeftRight;
      procedure DrawRightLeft;
      function GetItemAtCursor(X: Integer): Integer;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
   published
      property Active: Boolean read GetActive write SetActive;
      property Align;
      property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle;
      property Color;
      property Cursor;
      property DragCursor;
      property DragKind;
      property DragMode;
      property Direction: TScrollDirection read FDirection write SetDirection;
      property Enabled;
      property Height;
      property Hint;
      property Items: TScrollerCollection read FItems write SetItems;
      property Options: TScrollerOptions read FOptions write SetOptions;
      property ParentColor;
      property ParentFont;
      property ParentShowHint;
      property PopupMenu;
      property ScrollSpeed: TScrollSpeed read FSpeed write SetSpeed;
      property ShowHint;
      property Width;
      property Visible;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDock;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnHyperlinkClick: THyperlinkClick read FOnHyperlinkClick write FOnHyperlinkClick;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDock;
      property OnStartDrag;
   end;

const
  crScrollHand = TCursor(-23);
  crScrollHandHold = TCursor(-24);

procedure Register;

implementation

{$R *.RES}

{ TScrollerItem }

constructor TScrollerItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FImg := TBitmap.Create;
  FTextImg := TBitmap.Create;
  FUseText := True;
  FMargin := 4;
  FBkColor := clBlack;  
  FFont := TFont.Create;
  with FFont do
  begin
    OnChange := FontChange;
    Name := 'MS Sans Serif';
    Size := 8;
    Color := clWhite;
  end;
end;

destructor TScrollerItem.Destroy;
begin
  FImg.Free;
  FTextImg.Free;
  FFont.Free;
  inherited Destroy;
end;

procedure TScrollerItem.Change;
var
  GodParent: TGsScroller;
begin
  GodParent := TGsScroller(TScrollerCollection(Collection).GetOwner);
  if not GodParent.Active then GodParent.Invalidate;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TScrollerItem.Repaint;
var
  GodParent: TGsScroller;
begin
  if FUseText then DrawTextBitmap;
  GodParent := TGsScroller(TScrollerCollection(Collection).GetOwner);
  if not GodParent.Active then GodParent.Invalidate;
end;

function TScrollerItem.GetFace: TBitmap;
begin
  if FUseText then Result := FTextImg else Result := FImg;
end;

procedure TScrollerItem.DrawTextBitmap;
var
  GodParent: TGsScroller;
  R: TRect;
  S: string;
  Attr: Word;
begin
  GodParent := TGsScroller(TScrollerCollection(Collection).GetOwner);
  FTextImg.Height := GodParent.ClientHeight;
  FTextImg.Canvas.Brush.Color := FBkColor;
  FTextImg.Canvas.Font.Assign(FFont);
  if (FText1 = '') and (FText2 = '') then
  begin
    if FMargin = 0 then
      FTextImg.Width := 1
    else
      FTextImg.Width := FMargin;
    R := Rect(0, 0, FTextImg.Width, FTextImg.Height);
    FTextImg.Canvas.FillRect(R);
    Exit;
  end;
  Attr := DT_SINGLELINE or DT_NOPREFIX;
  case FAlign of
    taLeftJustify: Attr := Attr or DT_LEFT;
    taRightJustify: Attr := Attr or DT_RIGHT;
    taCenter: Attr := Attr or DT_CENTER;
  end;
  if ((FText1 = '') and (FText2 <> '')) or ((FText2 = '') and (FText1 <> '')) then
  begin
    if ((FText1 = '') and (FText2 <> '')) then S := FText2 else S := FText1;
    R := Rect(0, 0, FTextImg.Width, FTextImg.Height);
    FTextImg.Canvas.FillRect(R);
    Attr := Attr or DT_VCENTER;
    FTextImg.Width := FTextImg.Canvas.TextWidth(S) + (FMargin * 2);
    R := Rect(FMargin, 0, FTextImg.Width - FMargin, FTextImg.Height);
    DrawText(FTextImg.Canvas.Handle, PChar(S), Length(S), R, Attr)
  end else
  begin
    if FTextImg.Canvas.TextWidth(FText1) > FTextImg.Canvas.TextWidth(FText2) then
      FTextImg.Width := FTextImg.Canvas.TextWidth(FText1) + (FMargin * 2)
    else
      FTextImg.Width := FTextImg.Canvas.TextWidth(FText2) + (FMargin * 2);
    R := Rect(0, 0, FTextImg.Width, FTextImg.Height);
    FTextImg.Canvas.FillRect(R);
    R := Rect(FMargin, 0, FTextImg.Width - FMargin, FTextImg.Height);
    DrawText(FTextImg.Canvas.Handle, PChar(FText1), Length(FText1), R, Attr);
    R.Top := R.Top + FTextImg.Canvas.TextHeight(FText1);
    R.Bottom := R.Bottom + FTextImg.Canvas.TextHeight(FText1);
    DrawText(FTextImg.Canvas.Handle, PChar(FText2), Length(FText2), R, Attr);
  end;
  if FUseText then Change;
end;

procedure TScrollerItem.FontChange(Sender: TObject);
begin
  if FUseText then DrawTextBitmap;
end;

procedure TScrollerItem.SetImg(Value: TBitmap);
begin
  FImg.Assign(Value);
  if not FImg.Empty then FUseText := False;
  Change;
end;

procedure TScrollerItem.SetText1(Value: TCaption);
begin
  FText1 := Value;
  DrawTextBitmap;
  Change;
end;

procedure TScrollerItem.SetText2(Value: TCaption);
begin
  FText2 := Value;
  DrawTextBitmap;
  Change;
end;

procedure TScrollerItem.SetAlign(Value: TAlignment);
begin
  FAlign := Value;
  DrawTextBitmap;
  Change;
end;

procedure TScrollerItem.SetUseText(Value: Boolean);
begin
  FUseText := Value;
  if FUseText then DrawTextBitmap;
  Change;
end;

procedure TScrollerItem.SetMargin(Value: Word);
begin
  if Value > 8 then Value := 8;
  FMargin := Value;
  DrawTextBitmap;
  Change;
end;

procedure TScrollerItem.SetBkColor(Value: TColor);
begin
  FBkColor := Value;
  DrawTextBitmap;
  Change;
end;

procedure TScrollerItem.Assign(Source: TPersistent);
var
  SI: TScrollerItem;
begin
  if Source is TScrollerItem then
  begin
    SI := TScrollerItem(Source);
    FAlign := SI.Alignment;
    FBkColor := SI.Background;
    FFont.Assign(SI.Font);
    FHyperlink := SI.Hyperlink;
    FImg.Assign(SI.Bitmap);
    FText1 := SI.Text1;
    FText2 := Si.Text2;
    FUseText := SI.UseTextImage;
    if FUseText then DrawTextBitmap;
    Change;
  end else
    inherited Assign(Source);
end;

{ TScrollerCollection }

constructor TScrollerCollection.Create(Owner: TGsScroller);
begin
  inherited Create(TScrollerItem);
  FOwner := Owner;
end;

function TScrollerCollection.Add: TScrollerItem;
begin
  Result := TScrollerItem(inherited Add);
end;

function TScrollerCollection.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TScrollerCollection.GetItem(Index: Integer): TScrollerItem;
begin
  Result := TScrollerItem(inherited GetItem(Index));
end;

procedure TScrollerCollection.SetItem(Index: Integer; Value: TScrollerItem);
begin
  inherited SetItem(Index, Value);
end;

{ TGsScroller }

constructor TGsScroller.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := 240;
  Height := 32;
  Color := clBlack;
  ParentColor := False;
  FBorderStyle := bsSingle;
  FItems := TScrollerCollection.Create(Self);
  FTimer := TTimer.Create(Self);
  FTimer.OnTimer := ScrollTime;
  FTimer.Enabled := False;
  FSpeed := speed3_Normal;
  FTimer.Interval := 100;
  IncLeft := 0;
  IncRight := Width;
  LastLeft := -1;
  LastRight := -1;
  FIsMouseDown := False;
  FOptions := [opActiveMouse, opDrawFocus];
  FCursor := LoadCursor(hInstance, 'SCROLLHAND');
  Screen.Cursors[crScrollHand] := FCursor;
  FCursorHold := LoadCursor(hInstance, 'SCROLLHANDHLD');
  Screen.Cursors[crScrollHandHold] := FCursorHold;
end;

destructor TGsScroller.Destroy;
begin
  FItems.Free;
  FTimer.Free;
  inherited Destroy;
end;

procedure TGsScroller.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle] or WS_CLIPCHILDREN;
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    WindowClass.Style := WindowClass.Style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TGsScroller.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TGsScroller.SetDirection(Value: TScrollDirection);
begin
  if FDirection <> Value then
  begin
    FDirection := Value;
  end;
end;

function TGsScroller.GetActive: Boolean;
begin
  result := FTimer.Enabled;
end;

procedure TGsScroller.SetActive(Value: Boolean);
begin
  FTimer.Enabled := Value;
end;

procedure TGsScroller.SetSpeed(Value: TScrollSpeed);
begin
  FSpeed := Value;
  case FSpeed of
    speed1_VerySlow: FTimer.Interval := 110;
    else FTimer.Interval := 50;
  end;
end;

procedure TGsScroller.SetItems(Value: TScrollerCollection);
begin
  FItems.Assign(Value);
  Invalidate;
end;

procedure TGsScroller.SetOptions(Value: TScrollerOptions);
begin
  if FOptions <> Value then
  begin
    FOptions := Value;
    if (opActiveMouse in FOptions) then
    begin
      if Cursor <> crScrollHand then Cursor := crScrollHand;
    end else
      if Cursor = crScrollHand then Cursor := crDefault;
    Invalidate;
  end;
end;

procedure TGsScroller.Resize;
begin
  inherited Resize;
  Invalidate;
end;

procedure TGsScroller.Paint;
var
  R: TRect;
begin
  R := ClientRect;
  Canvas.Brush.Color := Self.Color;
  Canvas.FillRect(R);
  if FItems.Count = 0 then Exit;
  case FSpeed of
    speed1_VerySlow: FTimer.Interval := 110;
    else FTimer.Interval := 50;
  end;
  if (LastLeft < 0) or (LastLeft > pred(FItems.Count)) then LastLeft := 0;
  if (LastRight < 0) or (LastRight > pred(FItems.Count)) then LastRight := pred(FItems.Count);
  if FDirection = sdRightLeft then DrawRightLeft else DrawLeftRight;
end;

procedure TGsScroller.ScrollTime(Sender: TObject);
begin
  if FDirection = sdRightLeft then
    DrawRightLeft
  else
    DrawLeftRight;
end;

procedure TGsScroller.DrawRightLeft;
var
  I, L, StartI, W: Integer;
  B, Buffer: TBitmap;
  R: TRect;
label
  doAgain, doAgainDirectDraw;
begin
  if (FItems.Count = 0) or (LastLeft < 0) or (LastLeft > pred(FItems.Count)) then Exit;
  StartI := LastLeft;
  B := FItems[StartI].GetFace;
  W := B.Width;
  if (IncLeft + W) < 0 then
  begin
    IncLeft := IncLeft + W;
    LastLeft := LastLeft + 1;
    if LastLeft > pred(FItems.Count) then LastLeft := 0;
    StartI := LastLeft;
  end;
  L := IncLeft;

  if not (opDirectDraw in FOptions) then
  begin
    R := ClientRect;
    Buffer := TBitmap.Create;
    Buffer.Width := Width;
    Buffer.Height := Height;
    Buffer.Canvas.Brush.Color := Self.Color;
    Buffer.Canvas.FillRect(R);

    doAgain:

    for I := StartI to pred(FItems.Count) do
    begin
      B := FItems[I].GetFace;
      W := B.Width;
      Buffer.Canvas.Draw(L, 0, B);
      L := L + W;
      if not (csDesigning in ComponentState) then Application.ProcessMessages;
      if L > Width then
      begin
        LastRight := I;
        IncRight := L;
        Break;
      end;  
    end;
    if (L = 0) and (LastLeft = 0) then Exit; // All items are not visible!!!
    if L < Width then
    begin
      StartI := 0;
      goto doAgain;
    end;
    Canvas.Draw(0, 0, Buffer);
    Buffer.Free;
  end else
  begin
    doAgainDirectDraw:

    for I := StartI to pred(FItems.Count) do
    begin
      B := FItems[I].GetFace;
      W := B.Width;
      Canvas.Draw(L, 0, B);
      L := L + W;
      if not (csDesigning in ComponentState) then Application.ProcessMessages;
      if L > Width then
      begin
        LastRight := I;
        IncRight := L;
        Break;
      end;
    end;
    if (L = 0) and (LastLeft = 0) then Exit; // All items are not visible!!!
    if L < Width then
    begin
      StartI := 0;
      goto doAgainDirectDraw;
    end;
  end;

  if Active then
    case FSpeed of
      speed5_VeryFast: IncLeft := IncLeft - 4;
      speed4_Fast: IncLeft := IncLeft - 3;
      speed3_Normal: IncLeft := IncLeft - 2;
      else IncLeft := IncLeft - 1;
    end;
end;

procedure TGsScroller.DrawLeftRight;
var
  I, L, StartI, W: Integer;
  B, Buffer: TBitmap;
  R: TRect;
label
  doAgain, doAgainDirectDraw;
begin
  if (FItems.Count = 0) or (LastLeft < 0) or (LastLeft > pred(FItems.Count)) then Exit;
  StartI := LastRight;
  B := FItems[StartI].GetFace;
  W := B.Width;
  if IncRight < Width then IncRight := Width;
  if (IncRight - W) > Width then
  begin
    LastRight := LastRight - 1;
    if LastRight < 0 then LastRight := pred(FItems.Count);
    IncRight := Width + ((IncRight - W) - Width);
    StartI := LastRight;
  end;
  L := IncRight;

  if not (opDirectDraw in FOptions) then
  begin
    R := ClientRect;
    Buffer := TBitmap.Create;
    Buffer.Width := Width;
    Buffer.Height := Height;
    Buffer.Canvas.Brush.Color := Self.Color;
    Buffer.Canvas.FillRect(R);

    doAgain:

    for I := StartI downto 0 do
    begin
      B := FItems[I].GetFace;
      W := B.Width;
      Buffer.Canvas.Draw(L - W, 0, B);
      L := L - W;
      if not (csDesigning in ComponentState) then Application.ProcessMessages;
      if L < 0 then
      begin
        LastLeft := I;
        IncLeft := L;
        Break;
      end;
    end;
    if (L = Width) and (LastLeft = 0) then Exit; // All items are not visible!!!
    if L > 0 then
    begin
      StartI := pred(FItems.Count);
      goto doAgain;
    end;
    Canvas.Draw(0, 0, Buffer);
    Buffer.Free;
  end else
  begin

    doAgainDirectDraw:

    for I := StartI downto 0 do
    begin
      B := FItems[I].GetFace;
      W := B.Width;
      Canvas.Draw(L - W, 0, B);
      L := L - W;
      if not (csDesigning in ComponentState) then Application.ProcessMessages;
      if L < 0 then
      begin
        LastLeft := I;
        IncLeft := L;
        Break;
      end;
    end;
    if (L = Width) and (LastLeft = 0) then Exit; // All items are not visible!!!
    if L > 0 then
    begin
      StartI := pred(FItems.Count);
      goto doAgainDirectDraw;
    end;
  end;

  if Active then
    case FSpeed of
      speed5_VeryFast: IncRight := IncRight + 4;
      speed4_Fast: IncRight := IncRight + 3;
      speed3_Normal: IncRight := IncRight + 2;
      else IncRight := IncRight + 1;
    end;
end;

function TGsScroller.GetItemAtCursor(X: Integer): Integer;
begin
  FMouseItm := LastLeft;
  FMouseItmLeft := IncLeft + FItems[FMouseItm].GetFace.Width;
  while FMouseItmLeft < X do
  begin
    Inc(FMouseItm);
    if FMouseItm > pred(FItems.Count) then FMouseItm := 0;
    FMouseItmLeft := FMouseItmLeft + FItems[FMouseItm].GetFace.Width;
  end;
  FMouseItmLeft := FMouseItmLeft - FItems[FMouseItm].GetFace.Width;
  FMouseDifX := X - FMouseItmLeft;
  Result := FMouseItm;
end;

procedure TGsScroller.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if not Active or not (opActiveMouse in FOptions) or (Button <> mbLeft) then Exit;
  FIsMouseDown := True;
  Active := False;
  FMouseX := X;
  GetItemAtCursor(X);
  if (opDrawFocus in FOptions) then
  begin
    Canvas.Brush.Color := clWindow;
    Canvas.Pen.Color := clWindowText;
    Canvas.DrawFocusRect(Rect(FMouseItmLeft, 0,
      FMouseItmLeft + FItems[FMouseItm].GetFace.Width, ClientHeight));
    Canvas.Brush.Color := Color;
  end;
  Cursor := crScrollHandHold;
  Screen.Cursor := crScrollHandHold;
end;

procedure TGsScroller.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  L, Itm, W, StartI, KeepI: Integer;
  B: TBitmap;
label
  GoDown, GoUp;
begin
  inherited MouseMove(Shift, X, Y);
  if not FIsMouseDown or not (opActiveMouse in FOptions) then Exit;
  L := X - FMouseDifX;
  if FMouseItm > 0 then
    StartI := FMouseItm - 1
  else
    StartI := pred(FItems.Count);
  KeepI := StartI;
  GoDown:

  for Itm := StartI downto 0 do
  begin
    B := FItems[Itm].GetFace;
    W := B.Width;
    Canvas.Draw(L - W, 0, B);
    L := L - W;
    if L < 0 then
    begin
      LastLeft := Itm;
      IncLeft := L;
      Break;
    end;
    KeepI := Itm;
  end;
  if L > 0 then
  begin
    StartI := KeepI - 1;
    if StartI < 0 then StartI := pred(FItems.Count);
    goto GoDown;
  end;

  L := X - FMouseDifX;
  StartI := FMouseItm;
  KeepI := StartI;

  GoUp:

  for Itm := StartI to pred(FItems.Count) do
  begin
    B := FItems[Itm].GetFace;
    W := B.Width;
    Canvas.Draw(L, 0, B);
    L := L + W;
    if L > Width then
    begin
      LastRight := Itm;
      IncRight := L;
      Break;
    end;
    KeepI := Itm;
  end;
  if L < Width then
  begin
    StartI := KeepI + 1;
    if StartI > pred(FItems.Count) then StartI := 0;
    goto GoUp;
  end;
end;

procedure TGsScroller.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  I: Integer;
  S: string;
  Execute: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  if Button <> mbLeft then Exit;
  if (opExecuteHyperlinks in FOptions) and (FMouseX = X) then
  begin
    I := GetItemAtCursor(X);
    S := Trim(FItems[I].Hyperlink);
    if S <> '' then
    begin
      Execute := True;
      if Assigned(FOnHyperlinkClick) then FOnHyperlinkClick(Self, S, Execute);
      if not Execute or (Trim(S) = '') then Exit;
      if (pos('http://', LowerCase(S)) = 1) or (pos('mailto:', LowerCase(S)) = 1) then
        ShellExecute(Handle, 'open', PChar(S), nil, nil, SW_SHOWNORMAL);
      if (pos('exec:', LowerCase(S)) = 1) then
      begin
        Delete(S, 1, 5);
        if S <> '' then
          ShellExecute(Handle, 'open', PChar(S), nil, PChar(ExtractFilePath(S)), SW_SHOWNORMAL);
      end;
    end;
  end;
  if not FIsMouseDown or not (opActiveMouse in FOptions) then Exit;
  FIsMouseDown := False;
  Cursor := crScrollHand;
  Screen.Cursor := crDefault;
  if FMouseX = X then
  begin
    Active := True;
    Exit;
  end;
  if FMouseX > X then
  begin
    if FDirection <> sdRightLeft then
    begin
      if (FMouseX - X) in [1..50] then SetSpeed(speed1_VerySlow)
      else if (FMouseX - X) in [51..100] then SetSpeed(speed2_Slow)
      else SetSpeed(speed3_Normal);
      FDirection := sdRightLeft;
      Active := True;
      Exit;
    end;
    if (FMouseX - X) in [1..50] then SetSpeed(speed1_VerySlow)
    else if (FMouseX - X) in [51..100] then SetSpeed(speed2_Slow)
    else if (FMouseX - X) in [101..150] then SetSpeed(speed3_Normal)
    else if (FMouseX - X) in [151..200] then SetSpeed(speed4_Fast)
    else SetSpeed(speed5_VeryFast);
  end;
  if FMouseX < X then
  begin
    if FDirection <> sdLeftRight then
    begin
      if (X - FMouseX) in [1..50] then SetSpeed(speed1_VerySlow)
      else if (X - FMouseX) in [51..100] then SetSpeed(speed2_Slow)
      else SetSpeed(speed3_Normal);
      FDirection := sdLeftRight;
      Active := True;
      Exit;
    end;
    if (X - FMouseX) in [1..50] then SetSpeed(speed1_VerySlow)
    else if (X - FMouseX) in [51..100] then SetSpeed(speed2_Slow)
    else if (X - FMouseX) in [101..150] then SetSpeed(speed3_Normal)
    else if (X - FMouseX) in [151..200] then SetSpeed(speed4_Fast)
    else SetSpeed(speed5_VeryFast);
  end;
  Active := True;
end;

{ Register components }

procedure Register;
begin
  RegisterComponents('G''s Additional', [TGsScroller]);
end;

end.
